<?php
/**
 * Created by PhpStorm.
 * User: b2wsra
 * Date: 26.10.2018
 * Time: 09:09
 */

namespace Codevision;


class LicenseManager {


	/**
	 * We may only have one single instance throught wordpress!
	 *
	 * @var LicenseManager
	 */
	private static $singleton;

	/**
	 * @var string
	 */
	private $key;

	/**
	 * @var string
	 */
	private $namespace;

	/**
	 * @var bool
	 */
	private $valid;

	private $api_server;
	private $auth_key;
	private $plugin_slug;
	private $license_info_field;
	private $license_key_name;
	private $plugin_base_file;
	private $license_view_slug;

	private $other_requirements = [];
	private $min_php = '7.0';
	private $min_wp = '4.9.1';

	/**
	 * @var array
	 */
	private $additional_data;
	/**
	 * @var array
	 */
	private $messages;

	/**
	 * @var TwigRenderer
	 */
	private $renderer;

	private $upgrade_message;


	public static function getInstance() {

		if ( self::$singleton ) {
			return self::$singleton;
		}

		self::$singleton = new LicenseManager();

		return self::$singleton;
	}

	public function admin_init() {
		add_action( 'admin_init', [ $this, 'handle_plugin_compatibility' ] );
	}

	public function init() {

		if ( ! is_admin() ) {
			return;
		}

		if ( ! $this->is_valid() ) {
			add_action( 'admin_notices', [ $this, 'missing_or_wrong_api_key' ] );
		}

		if ( $this->is_trial() ) {
			add_action( 'admin_notices', [ $this, 'trial_period_activated' ] );
		}

		add_filter( 'acf/save_post', [ $this, 'update_license_data' ], 20 );
		add_filter( 'acf/prepare_field/key=' . $this->license_info_field, [ $this, 'render_license_info' ] );
		add_action( 'plugins_loaded', [ $this, 'load_text_domain' ] );

		if ( $this->is_updateable() ) {

			$updater = \Puc_v4_Factory::buildUpdateChecker( $this->api_server . '/api/auto-update/version.json', $this->plugin_base_file, $this->plugin_slug );

			$updater->addQueryArgFilter( [ $this, 'update_license' ] );
			$updater->addHttpRequestArgFilter( [ $this, 'update_license' ] );

		}

		$this->renderer = new TwigRenderer( $this->plugin_slug );
	}

	private function is_updateable() {

		$last_check = get_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ) );

		if ( $last_check === 'available' ) {
			return true;
		}

		if ( $last_check === 'error' ) {
			return false;
		}

		if ( $last_check === 'not_available' ) {
			return false;
		}

		$response = wp_remote_get( $this->api_server . '/api/auto-update/version.json', [ 'body' => $this->update_license( [] ) ] );

		if ( $response instanceof \WP_Error ) {

			Helper::log( $response->get_error_message(), 'error', $this->plugin_slug );

			set_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ), 'error', 12 * HOUR_IN_SECONDS );

			return false;
		}

		$response_code = wp_remote_retrieve_response_code( $response );

		if ( $response_code !== 200 ) {

			set_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ), 'not_available', 12 * HOUR_IN_SECONDS );

			return false;
		}

		$content_type = wp_remote_retrieve_header( $response, 'content-type' );

		if ( strpos( $content_type, 'application/json' ) === false ) {

			set_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ), 'error', 12 * HOUR_IN_SECONDS );

			return false;
		}

		$data = json_decode( wp_remote_retrieve_body( $response ), true );

		if ( ! $data ) {

			set_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ), 'error', 12 * HOUR_IN_SECONDS );

			return false;
		}

		if ( ! array_key_exists( 'download_url', $data ) ) {

			set_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ), 'not_available', 12 * HOUR_IN_SECONDS );

			return false;
		}

		if ( ! strlen( $data['download_url'] ) ) {

			set_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ), 'not_available', 12 * HOUR_IN_SECONDS );

			return false;
		}

		set_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ), 'available', 12 * HOUR_IN_SECONDS );

		return true;
	}

	public function load_text_domain() {
		load_plugin_textdomain( 'codevision-wp-license-manager', false, __DIR__ . '/languages/' );
	}

	public function update_license( $args ) {

		$args['key']    = get_field( $this->license_key_name, 'options' );
		$args['domain'] = get_site_url();

		return $args;
	}

	private function key( $key ) {
		return $this->namespace . $key;
	}


	/**
	 * Get the key from the key field
	 * and reset the transients in case of missing key
	 */
	private function load_key() {

		if ( $this->key ) {
			return;
		}

		$key = get_field( $this->license_key_name, 'options' );

		if ( ! $key ) {

			$this->clear();

			return;
		}

		$this->key = $key;
	}

	/**
	 * Clear all Transients and the current key
	 */
	public function clear() {

		delete_transient( $this->key( Constants::TRIAL_TIMEOUT_TRANSIENT ) );
		delete_transient( $this->key( Constants::VERIFICATION_TIMEOUT_TRANSIENT ) );
		delete_transient( $this->key( Constants::VERIFICATION_TRANSIENT ) );
		delete_transient( $this->key( Constants::LAST_RESPONSE_TRANSIENT ) );
		delete_transient( $this->key( Constants::UPDATE_CHECK_TRANSIENT ) );

		$this->key = null;
	}

	public function handle_plugin_compatibility() {

		$this->upgrade_message = null;
		$error                 = self::verify_environment( $this->min_php, $this->min_wp);

		if ( ! $error ) {
			return true;
		}

		deactivate_plugins( plugin_basename( $this->plugin_base_file ) );

		$messages = [
			__( '<h3>Could not activate the plugin!</h3>', 'codevision-wp-license-manager' ),
			__( '<p>This Plugin requires at least:</p>', 'codevision-wp-license-manager' )
		];

		$this->upgrade_message = implode( '', array_merge( $messages, $error, $this->other_requirements ) );

		add_action( 'admin_notices', [ $this, 'upgrade_environment_required' ] );

		return false;
	}

	public static function verify_environment( $min_php, $min_wp ) {

		global $wp_version;

		$error = [];

		if ( version_compare( PHP_VERSION, $min_php, '<' ) ) {
			$error[] = sprintf(  '<p>PHP with Version %s ( You have installed: %s )</p>', $min_php, PHP_VERSION );
		}

		if ( version_compare( $wp_version, $min_wp, '<' ) ) {
			$error[] = sprintf( '<p>WordPress with Version %s ( You have installed: %s )</p>', $min_wp, $wp_version );
		}

		return $error;
	}

	public function upgrade_environment_required() {
		if ( ! $this->upgrade_message ) {
			return;
		}

		?>

        <div class="notice-error notice is-dismissible">
			<?php echo $this->upgrade_message ?>
        </div>

	<?php }

	/**
	 * Set the correct state for the given installation by verifying the
	 * stored key or generating a trial license
	 *
	 * @param string $min_php
	 * @param string $min_wp
	 */
	public function activate() {

		$this->clear();

		// Check current key
		$this->load_key();

		// The current key is valid
		if ( $this->key && $this->validate_key() ) {
			return;
		}

		// clear transient, because otherwise we will not revalidate a possible new trial key
		$this->clear();

		// Check if trial license can be aquired
		$response = $this->ask_trial_key();

		if ( ! $response ) {
			return;
		}

		// Update the Value for the license Key and verify it
		update_field( $this->license_key_name, $response['license-key'], 'option' );

		$this->key = $response['license-key'];

		$this->validate_key();
	}

	public function deactivate() {
		$this->clear();
	}

	public function is_trial() {
		return $this->get_trial_period() > 0;
	}

	public function get_trial_period() {

		$response = $this->get_last_response();

		if ( ! $response ) {
			return false;
		}

		if ( ! is_array( $response ) ) {
			return false;
		}

		if ( ! array_key_exists( 'LicenseData', $response ) ) {
			return false;
		}

		if ( $response['LicenseData']['type'] !== 'trial' ) {
			return false;
		}

		$end = new \DateTime( $response['LicenseData']['end_date'] );
		$now = new \DateTime();

		$diff = date_diff( $end, $now );

		$days = intval( $diff->format( '%a' ) );

		// Handle the case that we do not have a full day left
		if ( $days === 0 && intval( $diff->format( '%s' ) ) > 0 ) {
			return 1;
		}

		return $days;
	}

	public function is_valid() {

		if ( $this->valid ) {
			return true;
		}

		$this->load_key();

		if ( ! $this->key ) {
			return false;
		}

		return $this->validate_key();
	}

	public function get_last_response() {
		$response = get_transient( $this->key( Constants::LAST_RESPONSE_TRANSIENT ) );

		return $response;
	}

	/**
	 * Validates the key against the verify_key method if last validation is not older than 12 hours return the
	 * verification result
	 *
	 * @return bool
	 */
	private function validate_key() {

		$key_verification = get_transient( $this->key( Constants::VERIFICATION_TRANSIENT ) );

		if ( $this->key === $key_verification ) {

			$this->valid = true;

			return true;
		}

		$valid_key = $this->verify_key();

		if ( ! $valid_key ) {
			return false;
		}

		set_transient( $this->key( Constants::VERIFICATION_TRANSIENT ), $this->key, 12 * HOUR_IN_SECONDS );

		return true;
	}

	/**
	 * Use the License Server to verify the Key entered in the Key field
	 *
	 * @param string $action
	 *
	 * @return bool
	 */
	private function verify_key( $action = 'validate' ) {
		$verification_timeout = $this->key( Constants::VERIFICATION_TIMEOUT_TRANSIENT );

		try {

			if ( get_transient( $verification_timeout ) ) {
				return false;
			}

			set_transient( $verification_timeout, true, 5 * MINUTE_IN_SECONDS );

			delete_transient( $this->key( Constants::LAST_RESPONSE_TRANSIENT ) );

			if ( ! $this->additional_data ) {
				$this->additional_data = [];
			}

			$response = wp_remote_get( $this->api_server . '/api/' . $action, [
				'body' => array_merge( $this->additional_data, [
					'auth'   => $this->auth_key,
					'key'    => $this->key,
					'domain' => get_site_url()
				] )
			] );

			if ( $response instanceof \WP_Error ) {

				Helper::log( $response->get_error_message(), 'error', $this->plugin_slug );

				return false;
			}

			$body     = wp_remote_retrieve_body( $response );
			$response = json_decode( $body, true );

			// We have to activate the license here
			if ( $action === 'validate' && array_key_exists( 'ErrorCode', $response ) && $response['ErrorCode'] === 201 ) {

				delete_transient( $verification_timeout );

				return $this->verify_key( 'activate' );
			}

			if ( $response['Success'] === false ) {

				Helper::log( sprintf( 'Could not validate license: %s', json_encode( $response ) ), 'error', $this->plugin_slug );

				return false;
			}

			set_transient( $this->key( Constants::LAST_RESPONSE_TRANSIENT ), $response, 24 * HOUR_IN_SECONDS );

			$this->valid = true;

			return true;

		} catch ( \Exception $e ) {
			Helper::log( $e, 'error', $this->plugin_slug );
		}

		return false;

	}

	/**
	 * Ask the License Server to get a trial license, but only once every 5 Minutes
	 *
	 * @return bool
	 */
	private function ask_trial_key() {

		// Skip, if we already have a key (this could be an expired one, we don't care)
		if ( $this->key ) {
			return false;
		}

		$trial_timeout = $this->key( Constants::TRIAL_TIMEOUT_TRANSIENT );

		if ( get_transient( $trial_timeout ) ) {
			return false;
		}

		set_transient( $trial_timeout, true, 5 * MINUTE_IN_SECONDS );

		$response = wp_remote_get( $this->api_server . '/api/' . 'generate', [
			'body' => [
				'auth'   => $this->auth_key,
				'type'   => 'trial',
				'domain' => get_site_url()
			]
		] );

		if ( $response instanceof \WP_Error ) {

			Helper::log( $response->get_error_message(), 'error', $this->plugin_slug );

			return false;
		}

		$body     = wp_remote_retrieve_body( $response );
		$response = json_decode( $body, true );

		if ( $response['Success'] === false ) {

			Helper::log( sprintf( 'Could not activate trial license: %s', json_encode( $response ) ), 'error', $this->plugin_slug );

			return false;
		}

		return $response['LicenseData'];
	}


	public function render_license_info( $field ) {

		$response = $this->get_last_response();

		if ( ! $response ) {
			return $field;
		}

		if ( array_key_exists( 'DigistoreData', $response ) ) {

			$response = $response['DigistoreData'];

		} else {

			$day_format = 'd.m.Y';

			if ( $response['LicenseData']['type'] === 'trial' ) {
				$day_format = 'd.m.Y H:i';
			}

			$response['product_name'] = $response['LicenseData']['description'];

			$response['last_payment_at_msg'] = date( $day_format, strtotime( $response['LicenseData']['start_date'] ) );

			if ( $response['LicenseData']['end_date'] ) {
				$response['next_payment_at_msg'] = date( $day_format, strtotime( $response['LicenseData']['end_date'] ) );
				$response['paid_until_msg']      = date( $day_format, strtotime( $response['LicenseData']['end_date'] ) );
			}
		}

		$days_left = $this->get_trial_period();

		if ( $days_left > 0 ) {
			$response['trial'] = sprintf( $this->get_translated_text( 'trial' ), $days_left );
		}

		if ( $days_left === 1 ) {
			$response['trial'] = sprintf( $this->get_translated_text( 'trial_one_day_left' ) );
		}

		$field['message'] = $this->renderer->view( 'license_info', $response );

		return $field;
	}

	private function get_translated_text( $key ) {
		$msg = $this->messages[ $key ];

		if ( ! is_array( $msg ) ) {
			return __( $msg, $this->plugin_slug );
		}

		$text = array_shift( $msg );

		return vsprintf( __( $text, $this->plugin_slug ), $msg );
	}

	public function update_license_data() {
		$screen = get_current_screen();

		if ( strpos( $screen->id, $this->license_view_slug ) === false ) {
			return;
		}

		$this->activate();
	}

	public function missing_or_wrong_api_key() {

		$message = [
			'headline' => $this->get_translated_text( 'headline' ),
			'message'  => $this->get_translated_text( 'missing_license_admin_notice' )
		];

		echo $this->renderer->view( 'missing_license_admin_notice', $message );
	}

	public function trial_period_activated() {

		$hide_trial = $this->key( Constants::MESSAGE_TRIAL_TIMEOUT_TRANSIENT );

		if ( get_transient( $hide_trial ) ) {
			return;
		}

		$days_left = $this->get_trial_period();

		$message = [
			'headline' => $this->get_translated_text( 'headline' ),
			'message'  => sprintf( $this->get_translated_text( 'trial_period_admin_notice' ), $days_left )
		];

		// Show nag screen only every 2 hours
		set_transient( $hide_trial, true, 2 * HOUR_IN_SECONDS );

		echo $this->renderer->view( 'trial_period_admin_notice', $message );
	}

	public function setNamespace( $namespace ) {
		$this->namespace = $namespace;

		return $this;
	}

	public function setApiServer( $api_server ) {
		$this->api_server = $api_server;

		return $this;
	}

	public function setAuthKey( $auth_key ) {
		$this->auth_key = $auth_key;

		return $this;
	}

	public function setPluginSlug( $plugin_slug ) {
		$this->plugin_slug = $plugin_slug;

		return $this;
	}

	public function setLicenseInfoField( $license_info_field ) {
		$this->license_info_field = $license_info_field;

		return $this;
	}

	public function setLicenseKeyName( $license_key_name ) {
		$this->license_key_name = $license_key_name;

		return $this;
	}

	public function setPluginBasefile( $plugin_base_file ) {
		$this->plugin_base_file = $plugin_base_file;

		return $this;
	}

	public function setLicenseViewSlug( $license_view_slug ) {
		$this->license_view_slug = $license_view_slug;

		return $this;
	}

	public function setMessages( $messages ) {
		$this->messages = $messages;

		return $this;
	}

	public function setAdditionalData( $additional_data ) {
		$this->additional_data = $additional_data;

		return $this;
	}

	public function setMinPhp( $min_php ) {
		$this->min_php = $min_php;

		return $this;
	}

	public function setMinWp( $min_wp ) {
		$this->min_wp = $min_wp;

		return $this;
	}

	public function setOtherRequirements( $other_requirements ) {
		$this->other_requirements = $other_requirements;

		return $this;
	}

}